package ch.epfl.ppo;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingConstants;

public final class RPNCalc {
    private StringBuilder currentInput = new StringBuilder();
    private Stack<Double> stack = LinkedStack.emptyStack();
    private Stack<Stack<Double>> undoStack = LinkedStack.emptyStack();

    private JLabel topOfStackLabel = new JLabel();

    private String[] buttonLabels = new String[] {
            "7", "8", "9",     "/", "Clear",
            "4", "5", "6",     "*", "Clear All",
            "1", "2", "3",     "-", "Undo",
            "0", ".", "Enter", "+", "Redo"
    };
    private Map<String, JButton> buttons = new HashMap<>();

    private void setStackWithUndo(Stack<Double> newStack) {
        undoStack = undoStack.push(stack);
        stack = newStack;
    }

    private void setStackWithRedo(Stack<Double> newStack) {
        stack = newStack;
    }

    private void updateUI() {
        // Update button selection state
        boolean enoughElementsOnStack = (stack.size() >= 2)
                || (stack.size() == 1 && currentInput.length() > 0);
        for (String l: Arrays.asList("+", "-", "*", "/"))
            buttons.get(l).setEnabled(enoughElementsOnStack);

        boolean noDotInInput = currentInput.indexOf(".") == -1;
        buttons.get(".").setEnabled(noDotInInput);

        boolean canEnter = currentInput.length() > 0 || !stack.isEmpty();
        buttons.get("Enter").setEnabled(canEnter);

        boolean canUndo = !undoStack.isEmpty();
        buttons.get("Undo").setEnabled(canUndo);

        // Update top of stack label
        String newText;
        if (currentInput.length() != 0)
            newText = currentInput.toString();
        else if (!stack.isEmpty())
            newText = stack.top().toString();
        else
            newText = "empty stack";
        topOfStackLabel.setText(newText + "  ");
    }

    private void pushCurrentIfNonEmpty() {
        if (currentInput.length() > 0) {
            setStackWithUndo(stack.push(Double.parseDouble(currentInput.toString())));
            currentInput.setLength(0);
        }
    }

    private void start() {
        topOfStackLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        ActionListener buttonListener = new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                JButton b = (JButton)e.getSource();
                switch (b.getText()) {
                case "0": case "1": case "2": case "3": case "4":
                case "5": case "6": case "7": case "8": case "9":
                case ".":
                    currentInput.append(b.getText());
                    break;

                case "+": case "-": case "*": case "/":
                    pushCurrentIfNonEmpty();
                    double a2 = stack.top();
                    Stack<Double> s2 = stack.pop();
                    double a1 = s2.top();
                    double r = 0;
                    switch (b.getText()) {
                    case "+": r = a1 + a2; break;
                    case "-": r = a1 - a2; break;
                    case "*": r = a1 * a2; break;
                    case "/": r = a1 / a2; break;
                    }
                    setStackWithUndo(s2.pop().push(r));
                    break;

                case "Enter":
                    if (currentInput.length() > 0)
                        pushCurrentIfNonEmpty();
                    else
                        setStackWithUndo(stack.push(stack.top()));
                    break;

                case "Clear":
                    currentInput.setLength(0);
                    break;

                case "Clear All":
                    currentInput.setLength(0);
                    setStackWithUndo(LinkedStack.<Double>emptyStack());
                    break;

                case "Undo":
                    pushCurrentIfNonEmpty();
                    setStackWithRedo(undoStack.top());
                    undoStack = undoStack.pop();
                    break;

                case "Redo":
                    // TODO à faire
                    break;
                }
                updateUI();
            }
        };

        JPanel buttonsPanel = new JPanel(new GridLayout(4, 4));
        for (String l: buttonLabels) {
            JButton button = new JButton(l);
            button.addActionListener(buttonListener);
            buttons.put(l, button);
            buttonsPanel.add(button);
        }

        JFrame frame = new JFrame("RPN Calculator");
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        Container contentPane = frame.getContentPane();
        contentPane.setLayout(new BorderLayout());
        contentPane.add(topOfStackLabel, BorderLayout.PAGE_START);
        contentPane.add(buttonsPanel, BorderLayout.CENTER);

        updateUI();

        frame.pack();
        frame.setVisible(true);
    }

    public static void main(String[] args) {
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() { new RPNCalc().start(); }
        });
    }
}
